<?php

namespace MontonioPayment\Controller\Frontend;

use Doctrine\ORM\EntityManager;
use Enlight_Controller_Plugins_ViewRenderer_Bootstrap;
use Exception;
use Montonio\Structs\PaymentData;
use MontonioPayment\Services\CheckoutService;
use MontonioPayment\Services\ConfigService;
use MontonioPayment\Services\PaymentMethodService;
use MontonioPayment\Traits\LoggingTrait;
use Shopware\Components\CSRFWhitelistAware;
use Shopware\Components\Model\ModelManager;
use Shopware\Models\Order\Order;
use Shopware_Controllers_Frontend_Payment;

class MontonioPayment extends Shopware_Controllers_Frontend_Payment implements CSRFWhitelistAware
{
    use LoggingTrait;

    protected $paymentMethod = PaymentMethodService::MONTONIO_PAYMENT_METHOD_DIRECT_PAYMENT;
    /**
     * Creates the order during checkout and redirects to payment gateway
     *
     * @return void
     * @throws Exception
     */
    public function indexAction()
    {
        /** @var CheckoutService $checkoutService */
        $checkoutService = $this->get(CheckoutService::class);

        if (!$checkoutService->validateCredentials()) {
            $this->redirect([
                'controller' => 'checkout',
                'action' => 'shippingPayment',
                'montonio_payment_error' => '1',
            ]);
        } else {
            $paymentReference = 'MON--' . uniqid() . '--' . Shopware()->Session()->offsetGet('sessionId');
            Shopware()->Session()->offsetSet('montonioIsCheckout', 1);
            Shopware()->Session()->offsetSet('montonioPaymentReference', $paymentReference);


            $redirectUrl = $checkoutService->generateRedirectUrlFromBasket($paymentReference, $this->paymentMethod);
            if (empty($redirectUrl)) {
                $this->confirmErrorAction();
            }
            $this->debug('start checkout redirect', [
                'paymentReference' => $paymentReference,
                'url' => $redirectUrl,
            ]);
            Shopware()->Session()->offsetSet('montonioBank', '');
            $this->redirect($redirectUrl);
        }
    }

    public function cardsAction()
    {
        $this->paymentMethod = PaymentMethodService::MONTONIO_PAYMENT_METHOD_CARDS;
        $this->indexAction();
    }

    /**
     * Redirects to payment gateway from existing unpaid order
     *
     * @return void
     * @throws Exception
     */
    public function payOrderAction()
    {
        Shopware()->Session()->offsetSet('montonioIsCheckout', 0);
        $orderId = $this->request->get('orderId');
        $token = $this->request->get('token');
        if (empty($orderId)) {
            throw new Exception('order id is empty');
        }
        /** @var CheckoutService $checkoutService */
        $checkoutService = $this->get(CheckoutService::class);
        $calculatedToken = $checkoutService->calculateOrderToken($orderId);
        if ($calculatedToken !== $token) {
            throw new Exception('token does not match: ' . $calculatedToken . ' vs. ' . $token);
        }
        try {
            /** @var ModelManager $entityManager */
            $entityManager = $this->get('models');
            /** @var Order $order */
            $order = $entityManager->getRepository(Order::class)->find($orderId);
            $redirectUrl = $checkoutService->generateRedirectUrlFromOrder($order);
            $this->debug('start order redirect', [
                'orderId' => $orderId,
                'url' => $redirectUrl,
            ]);
            $this->redirect($redirectUrl);
        } catch (Exception $e) {
            $this->error('pay order error', [
                'error' => $e->getMessage(),
                'orderId' => $orderId,
                'url' => $redirectUrl,
            ]);
            $this->redirect([
                'controller' => 'index',
                'montonioError' => 1,
            ]);
        }
    }

    /**
     * Executed after redirect from gateway
     *
     * @return void
     * @throws Exception
     */
    public function confirmAction()
    {
        Shopware()->Session()->offsetSet('montonioBank', '');
        $token = $this->request->get('order-token');
        $this->debug('start confirm action', [
            'token' => $token,
        ]);
        if (empty($token)) {
            $this->confirmErrorAction();
            return;
        }
        /** @var ConfigService $configService */
        $configService = $this->get(ConfigService::class);
        $paymentsClient = $configService->getPaymentsClient();
        try {
            $paymentData = $paymentsClient->decodePaymentToken($token);
        } catch (Exception $e) {
            $this->error('could not decode payment token', ['token' => $token, 'error' => $e->getMessage(), 'trace' => $e->getTrace()]);
            $this->confirmErrorAction();
            return;
        }
        $this->debug('confirm action payment data', [
            'data' => $paymentData,
        ]);

        if (empty($paymentData) || $paymentData->paymentStatus !== PaymentData::STATUS_PAID || $paymentData->accessKey !== $configService->getAccessKey()) {
            $this->confirmErrorAction($paymentData);
            return;
        }


        /** @var EntityManager $entityManager */
        $entityManager = $this->get('models');
        /** @var Order $order */
        $order = $entityManager->getRepository(Order::class)->findOneBy(['transactionId' => $paymentData->merchantReference]);
        if (!$order) {
            $orderNumber = $this->saveOrder(
                $paymentData->merchantReference,
                $paymentData->merchantReference
            );
        }
        $this->setPaymentStatusFromPaymentTokenData($paymentData);
        $isCheckout = (bool)Shopware()->Session()->offsetGet('montonioIsCheckout');
        Shopware()->Session()->offsetSet('montonioIsCheckout', 0);
        $this->redirect(
            $isCheckout ?
                [
                    'controller' => 'checkout',
                    'action' => 'finish',
                ]
                :
                [
                    'controller' => 'account',
                    'action' => 'orders',
                ]
        );
    }


    /**
     * Webhook action
     *
     * @return void
     * @throws Exception
     */
    public function webhookAction()
    {
        /** @var Enlight_Controller_Plugins_ViewRenderer_Bootstrap $viewRendererPlugin */
        $viewRendererPlugin = $this->Front()->Plugins()->ViewRenderer();
        $viewRendererPlugin->setNoRender();
        if ($content = $this->request->getRawBody()) {
            $content = json_decode($content, true);
            if (!empty($content['orderToken'])) {
                /** @var ConfigService $configService */
                $configService = $this->get(ConfigService::class);
                $paymentsClient = $configService->getPaymentsClient();
                try {
                    $paymentData = $paymentsClient->decodePaymentToken($content['orderToken']);
                } catch (Exception $e) {
                    $this->error('could not decode payment token from webhook', ['token' => $content['orderToken'], 'error' => $e->getMessage(), 'trace' => $e->getTrace()]);
                    return;
                }
                $this->debug('webhook payment data', [
                    'data' => $paymentData,
                ]);

                $paymentReference = $paymentData->merchantReference;
                $isFromCheckout = strpos($paymentReference, 'MON--') !== false;

                /** @var EntityManager $entityManager */
                $entityManager = $this->get('models');
                $i = 0;
                while (true) {
                    if ($isFromCheckout) {
                        /** @var Order $order */
                        $order = $entityManager->getRepository(Order::class)->findOneBy(['transactionId' => $paymentReference]);
                    } else {
                        /** @var Order $order */
                        $order = $entityManager->getRepository(Order::class)->findOneBy(['number' => $paymentReference]);
                    }

                    if (!empty($order)) {
                        break;
                    }
                    sleep(1);
                    if ($i++ > 8) {
                        break;
                    }
                }

                if (empty($order) && $isFromCheckout) {
                    $this->error('webhook without order - try to create one', [
                        'data' => $paymentData,
                    ]);

                    $referenceParts = explode('--', $paymentReference);
                    $sessionId = $referenceParts[2];
                    /** @var Order $order */
                    $order = $entityManager->getRepository(Order::class)->findOneBy(['temporaryId' => $sessionId]);
                    if ($order) {
                        $orderModule = Shopware()->Modules()->Order();
                        $order->setNumber($orderModule->sGetOrderNumber());
                        $order->setTransactionId($paymentReference);
                        $entityManager->persist($order);
                        $entityManager->flush($order);
                        $orderModule->setOrderStatus($order->getId(), 0);
                    }
                }

                $this->setPaymentStatusFromPaymentTokenData($paymentData);
            }
        }


    }

    public function getWhitelistedCSRFActions()
    {
        return [
            'webhook',
        ];
    }

    /**
     * @param \stdClass $paymentData
     * @return void
     */
    protected function setPaymentStatusFromPaymentTokenData($paymentData)
    {
        $paymentReference = $paymentData->merchantReference;
        $isFromCheckout = strpos($paymentReference, 'MON--') !== false;

        /** @var EntityManager $entityManager */
        $entityManager = $this->get('models');
        try {
            if ($isFromCheckout) {
                /** @var Order $order */
                $order = $entityManager->getRepository(Order::class)->findOneBy(['transactionId' => $paymentReference]);
            } else {
                /** @var Order $order */
                $order = $entityManager->getRepository(Order::class)->findOneBy(['number' => $paymentReference]);
            }
            if (empty($order)) {
                throw new Exception('No order for orderNumber = ' . $paymentReference);
            }
        } catch (Exception $e) {
            $this->error('unable to get order for payment data: ' . $e->getMessage(), ['data' => $paymentData]);
            return;
        }
        $paymentStatus = $paymentData->paymentStatus === PaymentData::STATUS_PAID ? CheckoutService::PAYMENT_STATUS_PAID : CheckoutService::PAYMENT_STATUS_REVIEW;
        if (number_format($paymentData->grandTotal, 2) !== number_format($order->getInvoiceAmount(), 2)) {
            $this->error('payment amount difference', [
                'paid' => $paymentData->grandTotal,
                'currentOrderAmount' => $order->getInvoiceAmount(),
            ]);
            $paymentStatus = CheckoutService::PAYMENT_STATUS_REVIEW;
        }
        $this->debug('set order status from payment token data', ['orderNumber' => $order->getNumber(), 'orderId' => $order->getId(), 'status' => $paymentStatus]);
        Shopware()->Modules()->Order()->setPaymentStatus($order->getId(), $paymentStatus);
    }

    protected function confirmErrorAction($paymentData = null)
    {
        $isCheckout = (bool)Shopware()->Session()->offsetGet('montonioIsCheckout');

        $this->error('montonio payment failed', [
            'paymentData' => $paymentData,
            'isCheckout' => $isCheckout?1:0,
            'trace' => debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS),
        ]);

        if ($isCheckout) {
            $this->redirect([
                'controller' => 'checkout',
                'action' => 'shippingPayment',
                'montonio_payment_error' => '1',
            ]);
        } else {

            $this->redirect([
                'controller' => 'account',
                'action' => 'orders',
                'montonio_payment_error' => '1',
            ]);
        }

    }
}


