<?php

namespace MontonioPayment\Services;


use Doctrine\ORM\EntityManagerInterface;
use Montonio\Clients\PaymentsClient;
use Montonio\Structs\Address;
use Montonio\Structs\LineItem;
use Montonio\Structs\Payment;
use Montonio\Structs\PaymentData;
use Montonio\Structs\PaymentMethodOptions;
use MontonioPayment\Traits\LoggingTrait;
use Shopware\Models\Order\Order;

class CheckoutService
{
    const PAYMENT_STATUS_PAID = 12;
    const PAYMENT_STATUS_REFUNDED = 20;
    const PAYMENT_STATUS_CANCELLED = 35;
    const PAYMENT_STATUS_REVIEW = 21;

    use LoggingTrait;

    /**
     * @var ConfigService
     */
    private $configService;
    /**
     * @var EntityManagerInterface
     */
    private $modelManager;

    /**
     * @param ConfigService $configService
     * @param EntityManagerInterface $modelManager
     */
    public function __construct(ConfigService $configService, EntityManagerInterface $modelManager)
    {
        $this->configService = $configService;
        $this->modelManager = $modelManager;
    }

    public function generateRedirectUrlFromOrder(Order $order)
    {
        $lineItems = [];

        foreach ($order->getDetails() as $detail) {
            $lineItems[] = new LineItem([
                'name' => $detail->getArticleName(),
                'quantity' => $detail->getQuantity(),
                'finalPrice' => $detail->getPrice() * $detail->getQuantity(),
            ]);
        }


        $locale = strtolower(substr($order->getShop()->getLocale()->getLocale(), 0, 2));

        $paymentsClient = $this->configService->getPaymentsClient();

        $payment = (new Payment())
            ->setMethod(Payment::PAYMENT_METHOD_PAYMENT_INITIATION)
            ->setMethodDisplay('')
            ->setAmount(round($order->getInvoiceAmount(), 2))
            ->setCurrency($order->getCurrency())
            ->setMethodOptions(
                (new PaymentMethodOptions())
                    ->setPaymentDescription($this->configService->getPaymentDescription())
                    ->setPreferredCountry('DE')
                    ->setPreferredLocale($locale)
                    ->setPreferredProvider(Shopware()->Session()->get('montonioBank', ''))
            );


        $paymentData = (new PaymentData())
            ->setAccessKey($this->configService->getAccessKey())
            ->setGrandTotal(round($order->getInvoiceAmount(), 2))
            ->setCurrency($order->getCurrency())
            ->setMerchantReference($order->getNumber())
            ->setReturnUrl($this->configService->getReturnUrl())
            ->setNotificationUrl($this->configService->getWebhookUrl())
            ->setExp(time() + 600)
            ->setLocale($locale)
            ->setLineItems($lineItems)
            ->setPayment($payment);
        $this->debug('create payment intent', ['paymentData' => $paymentData->toArray()]);
        $montonioOrder = $paymentsClient->createOrder($paymentData);
        $this->debug('payment intent result', ['montonioOrder' => $montonioOrder]);

        if(empty($montonioOrder['paymentUrl'])){
            $this->error('payment intent creation failed', ['montonioOrder' => $montonioOrder]);
            return null;
        }

        return $montonioOrder['paymentUrl'];
    }

    public function generateRedirectUrlFromBasket($paymentReference, $paymentMethod = PaymentMethodService::MONTONIO_PAYMENT_METHOD_DIRECT_PAYMENT)
    {
        $lineItems = [];
        //TODO add shipping and other misc positions
        foreach ($this->getBasket()['content'] as $item) {
            $lineItems[] = [
                'name' => $item['articlename'],
                'quantity' => (float)$item['quantity'],
                'finalPrice' => $item["priceNumeric"] * $item['quantity'],
            ];
        }

        $locale = strtolower(substr(Shopware()->Shop()->getLocale()->getLocale(), 0, 2));

        $paymentsClient = $this->configService->getPaymentsClient();


        $userData = $this->getUser();
        if(empty($userData) || empty($userData['billingaddress'])){
            $this->error('user data not found', ['userData' => $userData]);
            return null;
        }
        $billingAddress = $userData['billingaddress'];

        $billingAddress = (new Address())
            ->setFirstName($billingAddress['firstname'])
            ->setLastName($billingAddress['lastname'])
            ->setAddressLine1($billingAddress['street'])
            ->setLocality($billingAddress['city'])
            ->setPostalCode($billingAddress['zipcode'])
            ->setCountry($userData['additional']['country']['countryiso'])
            ->setEmail($userData['additional']['user']['email']);

        $payment = (new Payment())
            ->setMethod($paymentMethod)
            ->setMethodDisplay('')
            ->setAmount(round($this->getAmount(), 2))
            ->setCurrency(Shopware()->Shop()->getCurrency()->getCurrency())
            ->setMethodOptions(
                (new PaymentMethodOptions())
                    ->setPaymentDescription($this->configService->getPaymentDescription())
                    ->setPreferredCountry('DE')
                    ->setPreferredLocale($locale)
                    ->setPreferredProvider(Shopware()->Session()->get('montonioBank', ''))
            );

        $paymentData = (new PaymentData())
            ->setAccessKey($this->configService->getAccessKey())
            ->setGrandTotal(round($this->getAmount(), 2))
            ->setCurrency(Shopware()->Shop()->getCurrency()->getCurrency())
            ->setMerchantReference($paymentReference)
            ->setReturnUrl($this->configService->getReturnUrl())
            ->setNotificationUrl($this->configService->getWebhookUrl())
            ->setExp(time() + 600)
            ->setLocale($locale)
            ->setLineItems($lineItems)
            ->setBillingAddress($billingAddress)
            ->setPayment($payment);

        $this->debug('create payment intent', ['paymentData' => $paymentData->toArray()]);
        $montonioOrder = $paymentsClient->createOrder($paymentData);
        $this->debug('payment intent result', ['montonioOrder' => $montonioOrder]);

        if(empty($montonioOrder['paymentUrl'])){
            $this->error('payment intent creation failed', ['montonioOrder' => $montonioOrder]);
            return null;
        }

        return $montonioOrder['paymentUrl'];
    }

    public function getAmount()
    {
        $user = $this->getUser();
        $basket = $this->getBasket();
        if (!empty($user['additional']['charge_vat'])) {
            return empty($basket['AmountWithTaxNumeric']) ? $basket['AmountNumeric'] : $basket['AmountWithTaxNumeric'];
        }

        return $basket['AmountNetNumeric'];
    }


    public function getUser()
    {
        if (!empty(Shopware()->Session()->sOrderVariables['sUserData'])) {
            return Shopware()->Session()->sOrderVariables['sUserData'];
        }

        return null;
    }

    public function getBasket()
    {
        if (!empty(Shopware()->Session()->sOrderVariables['sBasket'])) {
            return Shopware()->Session()->sOrderVariables['sBasket'];
        }

        return Shopware()->Modules()->Basket()->sGetBasket();
    }

    public function generatePayOrderUrl($orderId)
    {
        return $this->configService->completeUrl(Shopware()->Front()->Router()->assemble([
            'controller' => 'MontonioPayment',
            'module' => 'Frontend',
            'action' => 'payOrder',
            'orderId' => $orderId,
            'token' => $this->calculateOrderToken($orderId),
        ]));
    }

    public function calculateOrderToken($orderId)
    {
        return md5($orderId . '|' . $this->configService->getSecretKey());
    }

    public function validateCredentials()
    {
        $client = new PaymentsClient($this->configService->getAccessKey(), $this->configService->getSecretKey(), $this->configService->getEnvironment());
        $methods = $client->getPaymentMethods();
        return (empty($methods['statusCode']) || $methods['statusCode'] == 200);
    }
}
